@@ -1,12 +1,20 @@
-/* Copyright (c) 2011 Adam Jakubek, Rafał Gałczyński
+/* Copyright (c) 2011-2013 Adam Jakubek, Rafał Gałczyński
  * Released under the MIT license (see attached LICENSE file).
  */
 
 #include <Python.h>
 #include <structmember.h>
+#include "py23macros.h"
 
-staticforward PyTypeObject DLListNodeType;
-staticforward PyTypeObject DLListIteratorType;
+#ifndef PyVarObject_HEAD_INIT
+    #define PyVarObject_HEAD_INIT(type, size) \
+        PyObject_HEAD_INIT(type) size,
+#endif
+
+
+static PyTypeObject DLListType;
+static PyTypeObject DLListNodeType;
+static PyTypeObject DLListIteratorType;
 
 
 /* DLListNode */
@@ -75,6 +83,7 @@ static DLListNodeObject* dllistnode_create(PyObject* prev,
         ((DLListNodeObject*)next)->prev = (PyObject*)node;
     }
 
+    Py_DECREF(node->list_weakref);
     node->list_weakref = PyWeakref_NewRef(owner_list, NULL);
 
     return node;
@@ -97,15 +106,53 @@ static void dllistnode_delete(DLListNodeObject* node)
         next->prev = node->prev;
     }
 
+    node->prev = Py_None;
+    node->next = Py_None;
+
     Py_DECREF((PyObject*)node);
 }
 
+/* Convenience function for formatting list node to a string.
+ * Pass PyObject_Repr or PyObject_Str in the fmt_func argument. */
+static PyObject* dllistnode_to_string(DLListNodeObject* self,
+                                      reprfunc fmt_func,
+                                      const char* prefix,
+                                      const char* suffix)
+{
+    PyObject* str = NULL;
+    PyObject* tmp_str;
+
+    assert(fmt_func != NULL);
+
+    str = Py23String_FromString(prefix);
+    if (str == NULL)
+        goto str_alloc_error;
+
+    tmp_str = fmt_func(self->value);
+    if (tmp_str == NULL)
+        goto str_alloc_error;
+    Py23String_ConcatAndDel(&str, tmp_str);
+
+    tmp_str = Py23String_FromString(suffix);
+    if (tmp_str == NULL)
+        goto str_alloc_error;
+    Py23String_ConcatAndDel(&str, tmp_str);
+
+    return str;
+
+str_alloc_error:
+    Py_XDECREF(str);
+    PyErr_SetString(PyExc_RuntimeError, "Failed to create string");
+    return NULL;
+}
+
 static void dllistnode_dealloc(DLListNodeObject* self)
 {
+    Py_DECREF(self->list_weakref);
     Py_DECREF(self->value);
     Py_DECREF(Py_None);
 
-    self->ob_type->tp_free((PyObject*)self);
+    PyObject_Del((PyObject*)self);
 }
 
 static PyObject* dllistnode_new(PyTypeObject* type,
@@ -128,6 +175,7 @@ static PyObject* dllistnode_new(PyTypeObject* type,
     self->list_weakref = Py_None;
 
     Py_INCREF(self->value);
+    Py_INCREF(self->list_weakref);
 
     return (PyObject*)self;
 }
@@ -162,56 +210,12 @@ static PyObject* dllistnode_call(DLListNodeObject* self,
 
 static PyObject* dllistnode_repr(DLListNodeObject* self)
 {
-    PyObject* str = NULL;
-    PyObject* tmp_str;
-
-    str = PyString_FromString("<dllistnode(");
-    if (str == NULL)
-        goto str_alloc_error;
-
-    tmp_str = PyObject_Repr(self->value);
-    if (tmp_str == NULL)
-        goto str_alloc_error;
-    PyString_ConcatAndDel(&str, tmp_str);
-
-    tmp_str = PyString_FromString(")>");
-    if (tmp_str == NULL)
-        goto str_alloc_error;
-    PyString_ConcatAndDel(&str, tmp_str);
-
-    return str;
-
-str_alloc_error:
-    Py_XDECREF(str);
-    PyErr_SetString(PyExc_RuntimeError, "Failed to create string");
-    return NULL;
+    return dllistnode_to_string(self, PyObject_Repr, "<dllistnode(", ")>");
 }
 
 static PyObject* dllistnode_str(DLListNodeObject* self)
 {
-    PyObject* str = NULL;
-    PyObject* tmp_str;
-
-    str = PyString_FromString("dllistnode(");
-    if (str == NULL)
-        goto str_alloc_error;
-
-    tmp_str = PyObject_Str(self->value);
-    if (tmp_str == NULL)
-        goto str_alloc_error;
-    PyString_ConcatAndDel(&str, tmp_str);
-
-    tmp_str = PyString_FromString(")");
-    if (tmp_str == NULL)
-        goto str_alloc_error;
-    PyString_ConcatAndDel(&str, tmp_str);
-
-    return str;
-
-str_alloc_error:
-    Py_XDECREF(str);
-    PyErr_SetString(PyExc_RuntimeError, "Failed to create string");
-    return NULL;
+    return dllistnode_to_string(self, PyObject_Str, "dllistnode(", ")");
 }
 
 static PyMemberDef DLListNodeMembers[] =
@@ -227,8 +231,7 @@ static PyMemberDef DLListNodeMembers[] =
 
 static PyTypeObject DLListNodeType =
 {
-    PyObject_HEAD_INIT(NULL)
-    0,                              /* ob_size */
+    PyVarObject_HEAD_INIT(NULL, 0)
     "llist.dllistnode",             /* tp_name */
     sizeof(DLListNodeObject),       /* tp_basicsize */
     0,                              /* tp_itemsize */
@@ -288,8 +291,8 @@ static Py_ssize_t py_ssize_t_abs(Py_ssize_t x)
 }
 
 /* Convenience function for locating list nodes using index. */
-static DLListNodeObject* dllist_get_node_at(DLListObject* self,
-                                            Py_ssize_t index)
+static DLListNodeObject* dllist_get_node_internal(DLListObject* self,
+                                                  Py_ssize_t index)
 {
     Py_ssize_t i;
     Py_ssize_t middle = self->size / 2;
@@ -299,7 +302,7 @@ static DLListNodeObject* dllist_get_node_at(DLListObject* self,
 
     if (index >= self->size || index < 0)
     {
-        PyErr_SetString(PyExc_IndexError, "No such index");
+        PyErr_SetString(PyExc_IndexError, "Index out of range");
         return NULL;
     }
 
@@ -342,19 +345,50 @@ static DLListNodeObject* dllist_get_node_at(DLListObject* self,
             node = (DLListNodeObject*)node->prev;
     }
 
-    self->last_accessed_node = (PyObject*)node;
-    self->last_accessed_idx = index;
-
     return node;
 }
 
 /* Convenience function for extending (concatenating in-place)
  * the list with elements from a sequence. */
-static int dllist_extend(DLListObject* self, PyObject* sequence)
+static int dllist_extend_internal(DLListObject* self, PyObject* sequence)
 {
     Py_ssize_t i;
     Py_ssize_t sequence_len;
 
+    if (PyObject_TypeCheck(sequence, &DLListType))
+    {
+        /* Special path for extending with a DLList.
+         * It's not strictly required but it will maintain
+         * the last accessed item. */
+        PyObject* iter_node_obj = ((DLListObject*)sequence)->first;
+        PyObject* last_node_obj = self->last;
+
+        while (iter_node_obj != Py_None)
+        {
+            DLListNodeObject* iter_node = (DLListNodeObject*)iter_node_obj;
+            PyObject* new_node;
+
+            new_node = (PyObject*)dllistnode_create(
+                self->last, NULL, iter_node->value, (PyObject*)self);
+
+            if (self->first == Py_None)
+                self->first = new_node;
+            self->last = new_node;
+
+            if (iter_node_obj == last_node_obj)
+            {
+                /* This is needed to terminate loop if self == sequence. */
+                break;
+            }
+
+            iter_node_obj = iter_node->next;
+        }
+
+        self->size += ((DLListObject*)sequence)->size;
+
+        return 1;
+    }
+
     sequence_len = PySequence_Length(sequence);
     if (sequence_len == -1)
     {
@@ -404,29 +438,29 @@ static PyObject* dllist_to_string(DLListObject* self,
 
     if (self->first == Py_None)
     {
-        str = PyString_FromString("dllist()");
+        str = Py23String_FromString("dllist()");
         if (str == NULL)
             goto str_alloc_error;
         return str;
     }
 
-    str = PyString_FromString("dllist([");
+    str = Py23String_FromString("dllist([");
     if (str == NULL)
         goto str_alloc_error;
 
-    comma_str = PyString_FromString(", ");
+    comma_str = Py23String_FromString(", ");
     if (comma_str == NULL)
         goto str_alloc_error;
 
     while ((PyObject*)node != Py_None)
     {
         if (node != (DLListNodeObject*)self->first)
-            PyString_Concat(&str, comma_str);
+            Py23String_Concat(&str, comma_str);
 
         tmp_str = fmt_func(node->value);
         if (tmp_str == NULL)
             goto str_alloc_error;
-        PyString_ConcatAndDel(&str, tmp_str);
+        Py23String_ConcatAndDel(&str, tmp_str);
 
         node = (DLListNodeObject*)node->next;
     }
@@ -434,10 +468,10 @@ static PyObject* dllist_to_string(DLListObject* self,
     Py_DECREF(comma_str);
     comma_str = NULL;
 
-    tmp_str = PyString_FromString("])");
+    tmp_str = Py23String_FromString("])");
     if (tmp_str == NULL)
         goto str_alloc_error;
-    PyString_ConcatAndDel(&str, tmp_str);
+    Py23String_ConcatAndDel(&str, tmp_str);
 
     return str;
 
@@ -458,13 +492,13 @@ static void dllist_dealloc(DLListObject* self)
     while (node != Py_None)
     {
         PyObject* next_node = ((DLListNodeObject*)node)->next;
-        Py_DECREF(node);
+        dllistnode_delete((DLListNodeObject*)node);
         node = next_node;
     }
 
     Py_DECREF(Py_None);
 
-    self->ob_type->tp_free((PyObject*)self);
+    PyObject_Del((PyObject*)self);
 }
 
 static PyObject* dllist_new(PyTypeObject* type,
@@ -508,35 +542,36 @@ static int dllist_init(DLListObject* self, PyObject* args, PyObject* kwds)
         return -1;
     }
 
-    return dllist_extend(self, sequence) ? 0 : -1;
+    return dllist_extend_internal(self, sequence) ? 0 : -1;
 }
 
-static int dllist_compare(DLListObject* self, DLListObject* other)
+static PyObject* dllist_node_at(PyObject* self, PyObject* indexObject)
 {
-    DLListNodeObject* self_node = (DLListNodeObject*)self->first;
-    DLListNodeObject* other_node = (DLListNodeObject*)other->first;
-    int result = 0;
+    DLListNodeObject* node;
+    Py_ssize_t index;
 
-    while (result == 0)
+    if (!Py23Int_Check(indexObject))
     {
-        if ((PyObject*)self_node == Py_None ||
-            (PyObject*)other_node == Py_None)
-        {
-            if ((PyObject*)self_node == Py_None)
-                return ((PyObject*)other_node == Py_None) ? 0 : -1;
-            else
-                return ((PyObject*)other_node == Py_None) ? 1 : 0;
-        }
-
-        result = PyObject_Compare(self_node->value, other_node->value);
-        if (PyErr_Occurred())
-            return 0;
-
-        self_node = (DLListNodeObject*)self_node->next;
-        other_node = (DLListNodeObject*)other_node->next;
+        PyErr_SetString(PyExc_TypeError, "Index must be an integer");
+        return NULL;
     }
 
-    return result;
+    index = Py23Int_AsSsize_t(indexObject);
+
+    if (index < 0)
+        index = ((DLListObject*)self)->size + index;
+
+    node = dllist_get_node_internal((DLListObject*)self, index);
+    if (node != NULL)
+    {
+        /* update last accessed node */
+        ((DLListObject*)self)->last_accessed_node = (PyObject*)node;
+        ((DLListObject*)self)->last_accessed_idx = index;
+
+        Py_INCREF(node);
+    }
+
+    return (PyObject*)node;
 }
 
 static PyObject* dllist_repr(DLListObject* self)
@@ -549,6 +584,117 @@ static PyObject* dllist_str(DLListObject* self)
     return dllist_to_string(self, PyObject_Str);
 }
 
+static long dllist_hash(DLListObject* self)
+{
+    long hash = 0;
+    PyObject* iter_node_obj = self->first;
+
+    while (iter_node_obj != Py_None)
+    {
+        long obj_hash;
+        DLListNodeObject* iter_node = (DLListNodeObject*)iter_node_obj;
+
+        obj_hash = PyObject_Hash(iter_node->value);
+        if (obj_hash == -1)
+            return -1;
+
+        hash ^= obj_hash;
+        iter_node_obj = iter_node->next;
+    }
+
+    return hash;
+}
+
+static PyObject* dllist_richcompare(DLListObject* self,
+                                    DLListObject* other,
+                                    int op)
+{
+    DLListNodeObject* self_node;
+    DLListNodeObject* other_node;
+    int satisfied = 1;
+
+    if (!PyObject_TypeCheck(other, &DLListType))
+    {
+        Py_INCREF(Py_NotImplemented);
+        return Py_NotImplemented;
+    }
+
+    if (self == other &&
+        (op == Py_EQ || op == Py_LE || op == Py_GE))
+        Py_RETURN_TRUE;
+
+    if (self->size != other->size)
+    {
+        if (op == Py_EQ)
+            Py_RETURN_FALSE;
+        else if (op == Py_NE)
+            Py_RETURN_TRUE;
+    }
+
+    /* Scan through sequences' items as long as they are equal. */
+    self_node = (DLListNodeObject*)self->first;
+    other_node = (DLListNodeObject*)other->first;
+
+    while ((PyObject*)self_node != Py_None &&
+            (PyObject*)other_node != Py_None)
+    {
+        satisfied = PyObject_RichCompareBool(
+            self_node->value, other_node->value, Py_EQ);
+
+        if (satisfied == 0)
+            break;
+
+        if (satisfied == -1)
+            return NULL;
+
+        self_node = (DLListNodeObject*)self_node->next;
+        other_node = (DLListNodeObject*)other_node->next;
+    }
+
+    /* Compare last item */
+    if (satisfied)
+    {
+        /* At least one of operands has been fully traversed.
+         * Either self_node or other_node is equal to Py_None. */
+        switch (op)
+        {
+        case Py_EQ:
+            satisfied = (self_node == other_node);
+            break;
+        case Py_NE:
+            satisfied = (self_node != other_node);
+            break;
+        case Py_LT:
+            satisfied = ((PyObject*)other_node != Py_None);
+            break;
+        case Py_GT:
+            satisfied = ((PyObject*)self_node != Py_None);
+            break;
+        case Py_LE:
+            satisfied = ((PyObject*)self_node == Py_None);
+            break;
+        case Py_GE:
+            satisfied = ((PyObject*)other_node == Py_None);
+            break;
+        default:
+            assert(0 && "Invalid rich compare operator");
+            PyErr_SetString(PyExc_ValueError, "Invalid rich compare operator");
+            return NULL;
+        }
+    }
+    else if (op != Py_EQ)
+    {
+        /* Both nodes are valid, but not equal */
+        satisfied = PyObject_RichCompareBool(
+            self_node->value, other_node->value, op);
+    }
+
+    if (satisfied)
+        Py_RETURN_TRUE;
+    else
+        Py_RETURN_FALSE;
+}
+
 static PyObject* dllist_appendleft(DLListObject* self, PyObject* arg)
 {
     DLListNodeObject* new_node;
@@ -626,6 +772,13 @@ static PyObject* dllist_insert(DLListObject* self, PyObject* args)
             return NULL;
         }
 
+        if (((DLListNodeObject*)ref_node)->list_weakref == Py_None)
+        {
+            PyErr_SetString(PyExc_ValueError,
+                "dllistnode does not belong to a list");
+            return NULL;
+        }
+
         list_ref = PyWeakref_GetObject(
             ((DLListNodeObject*)ref_node)->list_weakref);
         if (list_ref != (PyObject*)self)
@@ -656,6 +809,119 @@ static PyObject* dllist_insert(DLListObject* self, PyObject* args)
     return (PyObject*)new_node;
 }
 
+static PyObject* dllist_extendleft(DLListObject* self, PyObject* sequence)
+{
+    Py_ssize_t i;
+    Py_ssize_t sequence_len;
+
+    if (PyObject_TypeCheck(sequence, &DLListType))
+    {
+        /* Special path for extending with a DLList.
+         * It's not strictly required but it will maintain
+         * the last accessed item. */
+        PyObject* iter_node_obj = ((DLListObject*)sequence)->first;
+        PyObject* last_node_obj = ((DLListObject*)sequence)->last;
+
+        while (iter_node_obj != Py_None)
+        {
+            DLListNodeObject* iter_node = (DLListNodeObject*)iter_node_obj;
+            PyObject* new_node;
+
+            new_node = (PyObject*)dllistnode_create(
+                NULL, self->first, iter_node->value, (PyObject*)self);
+
+            self->first = new_node;
+            if (self->last == Py_None)
+                self->last = new_node;
+
+            if (iter_node_obj == last_node_obj)
+            {
+                /* This is needed to terminate loop if self == sequence. */
+                break;
+            }
+
+            iter_node_obj = iter_node->next;
+        }
+
+        self->size += ((DLListObject*)sequence)->size;
+
+        /* update index of last accessed item */
+        if (self->last_accessed_idx >= 0)
+            self->last_accessed_idx += ((DLListObject*)sequence)->size;
+
+        Py_RETURN_NONE;
+    }
+
+    sequence_len = PySequence_Length(sequence);
+    if (sequence_len == -1)
+    {
+        PyErr_SetString(PyExc_ValueError, "Invalid sequence");
+        return NULL;
+    }
+
+    for (i = 0; i < sequence_len; ++i)
+    {
+        PyObject* item;
+        PyObject* new_node;
+
+        item = PySequence_GetItem(sequence, i);
+        if (item == NULL)
+        {
+            PyErr_SetString(PyExc_ValueError,
+                "Failed to get element from sequence");
+            return NULL;
+        }
+
+        new_node = (PyObject*)dllistnode_create(
+            NULL, self->first, item, (PyObject*)self);
+
+        self->first = new_node;
+        if (self->last == Py_None)
+            self->last = new_node;
+
+        ++self->size;
+
+        /* update index of last accessed item */
+        if (self->last_accessed_idx >= 0)
+            ++self->last_accessed_idx;
+
+        Py_DECREF(item);
+    }
+
+    Py_RETURN_NONE;
+}
+
+static PyObject* dllist_extendright(DLListObject* self, PyObject* arg)
+{
+    if (!dllist_extend_internal(self, arg))
+        return NULL;
+
+    Py_RETURN_NONE;
+}
+
+static PyObject* dllist_clear(DLListObject* self)
+{
+    PyObject* iter_node_obj = self->first;
+
+    while (iter_node_obj != Py_None)
+    {
+        DLListNodeObject* iter_node = (DLListNodeObject*)iter_node_obj;
+
+        iter_node_obj = iter_node->next;
+        dllistnode_delete(iter_node);
+    }
+
+    /* invalidate last accessed item */
+    self->last_accessed_node = Py_None;
+    self->last_accessed_idx = -1;
+
+    self->first = Py_None;
+    self->last = Py_None;
+    self->size = 0;
+
+    Py_RETURN_NONE;
+}
+
 static PyObject* dllist_popleft(DLListObject* self)
 {
     DLListNodeObject* del_node;
@@ -749,6 +1015,13 @@ static PyObject* dllist_remove(DLListObject* self, PyObject* arg)
 
     del_node = (DLListNodeObject*)arg;
 
+    if (del_node->list_weakref == Py_None)
+    {
+        PyErr_SetString(PyExc_ValueError,
+            "dllistnode does not belong to a list");
+        return NULL;
+    }
+
     list_ref = PyWeakref_GetObject(del_node->list_weakref);
     if (list_ref != (PyObject*)self)
     {
@@ -773,11 +1046,66 @@ static PyObject* dllist_remove(DLListObject* self, PyObject* arg)
     Py_INCREF(del_node->value);
     value = del_node->value;
 
+    /* unlink from parent list */
+    Py_DECREF(del_node->list_weakref);
+    Py_INCREF(Py_None);
+    del_node->list_weakref = Py_None;
+
     dllistnode_delete(del_node);
 
     return value;
 }
 
+static PyObject* dllist_rotate(DLListObject* self, PyObject* nObject)
+{
+    Py_ssize_t n;
+    Py_ssize_t split_idx;
+    Py_ssize_t n_mod;
+    DLListNodeObject* new_first;
+    DLListNodeObject* new_last;
+
+    if (self->size <= 1)
+        Py_RETURN_NONE;
+
+    if (!Py23Int_Check(nObject))
+    {
+        PyErr_SetString(PyExc_TypeError, "n must be an integer");
+        return NULL;
+    }
+
+    n = Py23Int_AsSsize_t(nObject);
+    n_mod = py_ssize_t_abs(n) % self->size;
+
+    if (n_mod == 0)
+        Py_RETURN_NONE; /* no-op */
+
+    if (n > 0)
+        split_idx = self->size - n_mod; /* rotate right */
+    else
+        split_idx = n_mod;  /* rotate left */
+
+    new_last = dllist_get_node_internal(self, split_idx - 1);
+    assert(new_last != NULL);
+    new_first = (DLListNodeObject*)new_last->next;
+
+    ((DLListNodeObject*)self->first)->prev = self->last;
+    ((DLListNodeObject*)self->last)->next = self->first;
+
+    new_first->prev = Py_None;
+    new_last->next = Py_None;
+
+    self->first = (PyObject*)new_first;
+    self->last = (PyObject*)new_last;
+
+    if (self->last_accessed_idx >= 0)
+    {
+        self->last_accessed_idx =
+            (self->last_accessed_idx + self->size - split_idx) % self->size;
+    }
+
+    Py_RETURN_NONE;
+}
+
 static PyObject* dllist_iter(PyObject* self)
 {
     PyObject* args;
@@ -812,14 +1140,71 @@ static Py_ssize_t dllist_len(PyObject* self)
     return list->size;
 }
 
+static PyObject* dllist_concat(PyObject* self, PyObject* other)
+{
+    DLListObject* new_list;
+
+    new_list = (DLListObject*)PyObject_CallObject(
+        (PyObject*)&DLListType, NULL);
+
+    if (!dllist_extend_internal(new_list, self) ||
+        !dllist_extend_internal(new_list, other))
+    {
+        Py_DECREF(new_list);
+        return NULL;
+    }
+
+    return (PyObject*)new_list;
+}
+
+static PyObject* dllist_inplace_concat(PyObject* self, PyObject* other)
+{
+    if (!dllist_extend_internal((DLListObject*)self, other))
+        return NULL;
+
+    Py_INCREF(self);
+    return self;
+}
+
+static PyObject* dllist_repeat(PyObject* self, Py_ssize_t count)
+{
+    DLListObject* new_list;
+    Py_ssize_t i;
+
+    new_list = (DLListObject*)PyObject_CallObject(
+        (PyObject*)&DLListType, NULL);
+
+    for (i = 0; i < count; ++i)
+    {
+        if (!dllist_extend_internal(new_list, self))
+        {
+            Py_DECREF(new_list);
+            return NULL;
+        }
+    }
+
+    return (PyObject*)new_list;
+}
+
 static PyObject* dllist_get_item(PyObject* self, Py_ssize_t index)
 {
     DLListNodeObject* node;
 
-    node = dllist_get_node_at((DLListObject*)self, index);
-    Py_XINCREF(node);
+    node = dllist_get_node_internal((DLListObject*)self, index);
+    if (node != NULL)
+    {
+        PyObject* value = node->value;
 
-    return (PyObject*)node;
+        Py_XINCREF(value);
+
+        /* update last accessed node */
+        ((DLListObject*)self)->last_accessed_node = (PyObject*)node;
+        ((DLListObject*)self)->last_accessed_idx = index;
+
+        return value;
+    }
+
+    return NULL;
 }
 
 static int dllist_set_item(PyObject* self, Py_ssize_t index, PyObject* val)
@@ -828,7 +1213,7 @@ static int dllist_set_item(PyObject* self, Py_ssize_t index, PyObject* val)
     DLListNodeObject* node;
     PyObject* oldval;
 
-    node = dllist_get_node_at(list, index);
+    node = dllist_get_node_internal(list, index);
     if (node == NULL)
         return -1;
 
@@ -866,6 +1251,10 @@ static int dllist_set_item(PyObject* self, Py_ssize_t index, PyObject* val)
     node->value = val;
     Py_DECREF(oldval);
 
+    /* update last accessed node */
+    list->last_accessed_node = (PyObject*)node;
+    list->last_accessed_idx = index;
+
     return 0;
 }
 
@@ -877,8 +1266,18 @@ static PyMethodDef DLListMethods[] =
       "Append element at the end of the list" },
     { "appendright", (PyCFunction)dllist_appendright, METH_O,
       "Append element at the end of the list" },
+    { "clear", (PyCFunction)dllist_clear, METH_NOARGS,
+      "Remove all elements from the list" },
+    { "extend", (PyCFunction)dllist_extendright, METH_O,
+      "Append elements from iterable at the right side of the list" },
+    { "extendleft", (PyCFunction)dllist_extendleft, METH_O,
+      "Append elements from iterable at the left side of the list" },
+    { "extendright", (PyCFunction)dllist_extendright, METH_O,
+      "Append elements from iterable at the right side of the list" },
     { "insert", (PyCFunction)dllist_insert, METH_VARARGS,
       "Inserts element before node" },
+    { "nodeat", (PyCFunction)dllist_node_at, METH_O,
+      "Return node at index" },
     { "popleft", (PyCFunction)dllist_popleft, METH_NOARGS,
       "Remove first element from the list and return it" },
     { "pop", (PyCFunction)dllist_popright, METH_NOARGS,
@@ -887,6 +1286,8 @@ static PyMethodDef DLListMethods[] =
       "Remove last element from the list and return it" },
     { "remove", (PyCFunction)dllist_remove, METH_O,
       "Remove element from the list" },
+    { "rotate", (PyCFunction)dllist_rotate, METH_O,
+      "Rotate the list n steps to the right" },
     { NULL },   /* sentinel */
 };
 
@@ -904,18 +1305,20 @@ static PyMemberDef DLListMembers[] =
 static PySequenceMethods DLListSequenceMethods[] =
 {
     dllist_len,                 /* sq_length */
-    0,                          /* sq_concat */
-    0,                          /* sq_repeat */
+    dllist_concat,              /* sq_concat */
+    dllist_repeat,              /* sq_repeat */
     dllist_get_item,            /* sq_item */
     0,                          /* sq_slice */
     dllist_set_item,            /* sq_ass_item */
-    0                           /* sq_ass_slice */
+    0,                          /* sq_ass_slice */
+    0,                          /* sq_contains */
+    dllist_inplace_concat,      /* sq_inplace_concat */
+    0,                          /* sq_inplace_repeat */
 };
 
 static PyTypeObject DLListType =
 {
-    PyObject_HEAD_INIT(NULL)
-    0,                          /* ob_size */
+    PyVarObject_HEAD_INIT(NULL, 0)
     "llist.dllist",             /* tp_name */
     sizeof(DLListObject),       /* tp_basicsize */
     0,                          /* tp_itemsize */
@@ -923,12 +1326,12 @@ static PyTypeObject DLListType =
     0,                          /* tp_print */
     0,                          /* tp_getattr */
     0,                          /* tp_setattr */
-    (cmpfunc)dllist_compare,    /* tp_compare */
+    0,                          /* tp_compare */
     (reprfunc)dllist_repr,      /* tp_repr */
     0,                          /* tp_as_number */
     DLListSequenceMethods,      /* tp_as_sequence */
     0,                          /* tp_as_mapping */
-    0,                          /* tp_hash */
+    (hashfunc)dllist_hash,      /* tp_hash */
     0,                          /* tp_call */
     (reprfunc)dllist_str,       /* tp_str */
     0,                          /* tp_getattro */
@@ -938,7 +1341,8 @@ static PyTypeObject DLListType =
     "Doubly linked list",       /* tp_doc */
     0,                          /* tp_traverse */
     0,                          /* tp_clear */
-    0,                          /* tp_richcompare */
+    (richcmpfunc)dllist_richcompare,
+                                /* tp_richcompare */
     offsetof(DLListObject, weakref_list),
                                 /* tp_weaklistoffset */
     dllist_iter,                /* tp_iter */
@@ -971,7 +1375,7 @@ static void dllistiterator_dealloc(DLListIteratorObject* self)
     Py_XDECREF(self->current_node);
     Py_DECREF(self->list);
 
-    self->ob_type->tp_free((PyObject*)self);
+    PyObject_Del((PyObject*)self);
 }
 
 static PyObject* dllistiterator_new(PyTypeObject* type,
@@ -1030,8 +1434,7 @@ static PyObject* dllistiterator_iternext(PyObject* self)
 
 static PyTypeObject DLListIteratorType =
 {
-    PyObject_HEAD_INIT(NULL)
-    0,                                  /* ob_size */
+    PyVarObject_HEAD_INIT(NULL, 0)
     "llist.dllistiterator",             /* tp_name */
     sizeof(DLListIteratorObject),       /* tp_basicsize */
     0,                                  /* tp_itemsize */
